#include "feat.h"
#include "random.h"

#include "uct/internal.h"
#include "uct/walk.h"
extern void uct_genmove_setup(struct uct *u, struct board *b, enum stone color);


FILE* featf = NULL;
FILE* searchfeatf = NULL;
float search_based_neural_net_feats[19*19];

int fi2coord(int fi) {
  int x = fi%19;
  int y = fi/19;
  return (y+1)*21+x+1;
}

// generate search-based features
void gen_search_feats(struct engine *engine, struct board* b, struct move m, float* feats) {
  // first get a random number of simulations to run on the board
  int sims = fast_random(10) + 1;
  struct uct *u = (struct uct*) (engine->data);
  uct_genmove_setup(u, b, m.color);

  int i = 0;
  int j = 0;
  
  for(j = 0; j < 19*19; j++)
    feats[j] = 0;

  for (i = 0; i < sims; i++) {
    uct_playout(u, b, m.color, u->t);  // this calls play_random_game which writes to the global array
    for(j = 0; j < 19*19; j++) {
      feats[j] = feats[j] + search_based_neural_net_feats[j];
    }
  }
  
  for(j = 0; j < 19*19; j++) {
    if(m.color == S_WHITE)
      feats[j] = (feats[j] / sims) * -1;
    else
      feats[j] = feats[j] / sims;
  }
  
}

void print_search_feats(float* feats) {
  int i = 0;
  for (i = 0; i < 19*19; i++)
    fprintf(searchfeatf, "%+0.6f%c", feats[i], (i == 19*19 - 1) ? '\n':' ');
}


void genfeats(struct board* b, struct move m, int* feats) {
  int i = 0;
  int flip;
  if (m.color == S_WHITE) flip = 1;
  else flip = 0;
  for (i = 0; i < 19*19; i++) {
    int at = fi2coord(i);
    int v = 0;
    int k = 0;
    int lib;
    //    fprintf(stderr, "%d:%d ", at, board_at(b, at));

    // at = m ?
    if (m.coord == at) v |= 1;
    k += 1;

    // legal(at)?
    if (!board_is_valid_play(b, m.color, at))
      v |= (1 << k);
    k += 1;

    // stone?
    if (board_at(b, at) == S_NONE) {
      v |= (4 << k); 
    }
    else if (board_at(b, at) == S_BLACK) {
      if (flip) v |= (1 << k);
      else v |= (2 << k);
    }
    else /*if (board_at(b, at) == S_WHITE)*/ {
      if (flip) v |= (2 << k);
      else v |= (1 << k);
    }
    k += 3;

    // liberties(at)?
    if (board_at(b, at) != S_NONE) {
      lib = board_group_info(b, group_at(b, at)).libs;
      if (lib == 1) v |= (1 << k);
      else if (lib == 2) v |= (2 << k);
      else if (lib == 3) v |= (4 << k);
      else if (lib >= 4) v |= (8 << k);
    }
    k += 4;

    if (b->last_move.coord == at) v |= (1 << k);
    if (b->last_move2.coord == at) v |= (2 << k);
    k += 2;

    feats[i] = v;
  }
}
void printfeats(struct board* b, struct move m, int* feats) {
  int i = 0;
  for (i = 0; i < 19*19; i++) 
    fprintf(featf, "%03x%c", feats[i], (i==19*19-1)?'\n':' ');
}

